@@ -117,11 +117,10 @@ Each call to ``parse()`` returns a dict has this form::
 import re
 import base64
 import json
-import xml.etree.cElementTree
 import logging
 
-from botocore.compat import six, XMLParseError
-from botocore.eventstream import EventStream
+from botocore.compat import six, ETree, XMLParseError
+from botocore.eventstream import EventStream, NoInitialResponseError
 
 from botocore.utils import parse_timestamp, merge_dicts, \
     is_json_value_header, lowercase_dict
@@ -141,7 +140,7 @@ class ResponseParserFactory(object):
         You can specify any kwargs that are allowed by a ResponseParser
         class.  There are currently two arguments:
 
-            * timestamp_parser - A callable that can parse a timetsamp string
+            * timestamp_parser - A callable that can parse a timestamp string
             * blob_parser - A callable that can parse a blob type
 
         """
@@ -236,6 +235,10 @@ class ResponseParser(object):
         if response['status_code'] >= 301:
             if self._is_generic_error_response(response):
                 parsed = self._do_generic_error_parse(response)
+            elif self._is_modeled_error_shape(shape):
+                parsed = self._do_modeled_error_parse(response, shape)
+                # We don't want to decorate the modeled fields with metadata
+                return parsed
             else:
                 parsed = self._do_error_parse(response, shape)
         else:
@@ -258,6 +261,9 @@ class ResponseParser(object):
             parsed['ResponseMetadata'] = response_metadata
         return parsed
 
+    def _is_modeled_error_shape(self, shape):
+        return shape is not None and shape.metadata.get('exception', False)
+
     def _is_generic_error_response(self, response):
         # There are times when a service will respond with a generic
         # error response such as:
@@ -271,6 +277,9 @@ class ResponseParser(object):
         # To prevent this case from happening we first need to check
         # whether or not this response looks like the generic response.
         if response['status_code'] >= 500:
+            if 'body' not in response or response['body'] is None:
+                return True
+
             body = response['body'].strip()
             return body.startswith(b'<html>') or not body
 
@@ -293,6 +302,10 @@ class ResponseParser(object):
         raise NotImplementedError(
             "%s._do_error_parse" % self.__class__.__name__)
 
+    def _do_modeled_error_parse(self, response, shape, parsed):
+        raise NotImplementedError(
+            "%s._do_modeled_error_parse" % self.__class__.__name__)
+
     def _parse_shape(self, shape, node):
         handler = getattr(self, '_handle_%s' % shape.type_name,
                           self._default_handle)
@@ -310,6 +323,11 @@ class ResponseParser(object):
     def _default_handle(self, shape, value):
         return value
 
+    def _create_event_stream(self, response, shape):
+        parser = self._event_stream_parser
+        name = response['context'].get('operation_name')
+        return EventStream(response['body'], shape, parser, name)
+
 
 class BaseXMLResponseParser(ResponseParser):
     def __init__(self, timestamp_parser=None, blob_parser=None):
@@ -354,6 +372,8 @@ class BaseXMLResponseParser(ResponseParser):
     def _handle_structure(self, shape, node):
         parsed = {}
         members = shape.members
+        if shape.metadata.get('exception', False):
+            node = self._get_error_root(node)
         xml_dict = self._build_name_to_xml_node(node)
         for member_name in members:
             member_shape = members[member_name]
@@ -378,6 +398,13 @@ class BaseXMLResponseParser(ResponseParser):
                     parsed[member_name] = attribs[location_name]
         return parsed
 
+    def _get_error_root(self, original_root):
+        if self._node_tag(original_root) == 'ErrorResponse':
+            for child in original_root:
+                if self._node_tag(child) == 'Error':
+                    return child
+        return original_root
+
     def _member_key_name(self, shape, member_name):
         # This method is needed because we have to special case flattened list
         # with a serialization name.  If this is the case we use the
@@ -418,20 +445,21 @@ class BaseXMLResponseParser(ResponseParser):
 
     def _parse_xml_string_to_dom(self, xml_string):
         try:
-            parser = xml.etree.cElementTree.XMLParser(
-                target=xml.etree.cElementTree.TreeBuilder(),
+            parser = ETree.XMLParser(
+                target=ETree.TreeBuilder(),
                 encoding=self.DEFAULT_ENCODING)
             parser.feed(xml_string)
             root = parser.close()
         except XMLParseError as e:
             raise ResponseParserError(
                 "Unable to parse response (%s), "
-                "invalid XML received:\n%s" % (e, xml_string))
+                "invalid XML received. Further retries may succeed:\n%s" %
+                (e, xml_string))
         return root
 
     def _replace_nodes(self, parsed):
         for key, value in parsed.items():
-            if value.getchildren():
+            if list(value):
                 sub_dict = self._build_name_to_xml_node(value)
                 parsed[key] = self._replace_nodes(sub_dict)
             else:
@@ -488,7 +516,13 @@ class QueryParser(BaseXMLResponseParser):
             parsed['ResponseMetadata'] = {'RequestId': parsed.pop('RequestId')}
         return parsed
 
+    def _do_modeled_error_parse(self, response, shape):
+        return self._parse_body_as_xml(response, shape, inject_metadata=False)
+
     def _do_parse(self, response, shape):
+        return self._parse_body_as_xml(response, shape, inject_metadata=True)
+
+    def _parse_body_as_xml(self, response, shape, inject_metadata=True):
         xml_contents = response['body']
         root = self._parse_xml_string_to_dom(xml_contents)
         parsed = {}
@@ -499,7 +533,8 @@ class QueryParser(BaseXMLResponseParser):
                     shape.serialization['resultWrapper'],
                     root)
             parsed = self._parse_shape(shape, start)
-        self._inject_response_metadata(root, parsed)
+        if inject_metadata:
+            self._inject_response_metadata(root, parsed)
         return parsed
 
     def _find_result_wrapped_shape(self, element_name, xml_root_node):
@@ -538,30 +573,43 @@ class EC2QueryParser(QueryParser):
         # This is different from QueryParser in that it's RequestID,
         # not RequestId
         original = super(EC2QueryParser, self)._do_error_parse(response, shape)
-        original['ResponseMetadata'] = {
-            'RequestId': original.pop('RequestID')
-        }
+        if 'RequestID' in original:
+            original['ResponseMetadata'] = {
+                'RequestId': original.pop('RequestID')
+            }
         return original
 
+    def _get_error_root(self, original_root):
+        for child in original_root:
+            if self._node_tag(child) == 'Errors':
+                for errors_child in child:
+                    if self._node_tag(errors_child) == 'Error':
+                        return errors_child
+        return original_root
+
 
 class BaseJSONParser(ResponseParser):
 
     def _handle_structure(self, shape, value):
-        member_shapes = shape.members
-        if value is None:
-            # If the comes across the wire as "null" (None in python),
-            # we should be returning this unchanged, instead of as an
-            # empty dict.
-            return None
         final_parsed = {}
-        for member_name in member_shapes:
-            member_shape = member_shapes[member_name]
-            json_name = member_shape.serialization.get('name', member_name)
-            raw_value = value.get(json_name)
-            if raw_value is not None:
-                final_parsed[member_name] = self._parse_shape(
-                    member_shapes[member_name],
-                    raw_value)
+        if shape.is_document_type:
+            final_parsed = value
+        else:
+            member_shapes = shape.members
+            if value is None:
+                # If the comes across the wire as "null" (None in python),
+                # we should be returning this unchanged, instead of as an
+                # empty dict.
+                return None
+            final_parsed = {}
+            for member_name in member_shapes:
+                member_shape = member_shapes[member_name]
+                json_name = member_shape.serialization.get('name', member_name)
+                raw_value = value.get(json_name)
+                if raw_value is not None:
+                    final_parsed[member_name] = self._parse_shape(
+                        member_shapes[member_name],
+                        raw_value)
         return final_parsed
 
     def _handle_map(self, shape, value):
@@ -625,20 +673,6 @@ class BaseJSONParser(ResponseParser):
             return { 'message': body }
 
 
-class JSONParser(BaseJSONParser):
-    """Response parse for the "json" protocol."""
-    def _do_parse(self, response, shape):
-        # The json.loads() gives us the primitive JSON types,
-        # but we need to traverse the parsed JSON data to convert
-        # to richer types (blobs, timestamps, etc.
-        parsed = {}
-        if shape is not None:
-            original_parsed = self._parse_body_as_json(response['body'])
-            parsed = self._parse_shape(shape, original_parsed)
-        self._inject_response_metadata(parsed, response['headers'])
-        return parsed
-
-
 class BaseEventStreamParser(ResponseParser):
 
     def _do_parse(self, response, shape):
@@ -655,12 +689,24 @@ class BaseEventStreamParser(ResponseParser):
         return final_parsed
 
     def _do_error_parse(self, response, shape):
-        error = {
-            'Error': {
-                'Code': response['headers'].get(':error-code', ''),
-                'Message': response['headers'].get(':error-message', ''),
+        exception_type = response['headers'].get(':exception-type')
+        exception_shape = shape.members.get(exception_type)
+        if exception_shape is not None:
+            original_parsed = self._initial_body_parse(response['body'])
+            body = self._parse_shape(exception_shape, original_parsed)
+            error = {
+                'Error': {
+                    'Code': exception_type,
+                    'Message': body.get('Message', body.get('message', ''))
+                }
+            }
+        else:
+            error = {
+                'Error': {
+                    'Code': response['headers'].get(':error-code', ''),
+                    'Message': response['headers'].get(':error-message', ''),
+                }
             }
-        }
         return error
 
     def _parse_payload(self, response, shape, member_shapes, final_parsed):
@@ -715,22 +761,69 @@ class EventStreamXMLParser(BaseEventStreamParser, BaseXMLResponseParser):
 
     def _initial_body_parse(self, xml_string):
         if not xml_string:
-            return xml.etree.cElementTree.Element('')
+            return ETree.Element('')
         return self._parse_xml_string_to_dom(xml_string)
 
 
+class JSONParser(BaseJSONParser):
+
+    EVENT_STREAM_PARSER_CLS = EventStreamJSONParser
+
+    """Response parser for the "json" protocol."""
+    def _do_parse(self, response, shape):
+        parsed = {}
+        if shape is not None:
+            event_name = shape.event_stream_name
+            if event_name:
+                parsed = self._handle_event_stream(response, shape, event_name)
+            else:
+                parsed = self._handle_json_body(response['body'], shape)
+        self._inject_response_metadata(parsed, response['headers'])
+        return parsed
+
+    def _do_modeled_error_parse(self, response, shape):
+        return self._handle_json_body(response['body'], shape)
+
+    def _handle_event_stream(self, response, shape, event_name):
+        event_stream_shape = shape.members[event_name]
+        event_stream = self._create_event_stream(response, event_stream_shape)
+        try:
+            event = event_stream.get_initial_response()
+        except NoInitialResponseError:
+            error_msg = 'First event was not of type initial-response'
+            raise ResponseParserError(error_msg)
+        parsed = self._handle_json_body(event.payload, shape)
+        parsed[event_name] = event_stream
+        return parsed
+
+    def _handle_json_body(self, raw_body, shape):
+        # The json.loads() gives us the primitive JSON types,
+        # but we need to traverse the parsed JSON data to convert
+        # to richer types (blobs, timestamps, etc.
+        parsed_json = self._parse_body_as_json(raw_body)
+        return self._parse_shape(shape, parsed_json)
+
+
 class BaseRestParser(ResponseParser):
 
     def _do_parse(self, response, shape):
         final_parsed = {}
         final_parsed['ResponseMetadata'] = self._populate_response_metadata(
             response)
+        self._add_modeled_parse(response, shape, final_parsed)
+        return final_parsed
+
+    def _add_modeled_parse(self, response, shape, final_parsed):
         if shape is None:
             return final_parsed
         member_shapes = shape.members
         self._parse_non_payload_attrs(response, shape,
                                       member_shapes, final_parsed)
         self._parse_payload(response, shape, member_shapes, final_parsed)
+
+    def _do_modeled_error_parse(self, response, shape):
+        final_parsed = {}
+        self._add_modeled_parse(response, shape, final_parsed)
         return final_parsed
 
     def _populate_response_metadata(self, response):
@@ -740,7 +833,7 @@ class BaseRestParser(ResponseParser):
             metadata['RequestId'] = headers['x-amzn-requestid']
         elif 'x-amz-request-id' in headers:
             metadata['RequestId'] = headers['x-amz-request-id']
-            # HostId is what it's called whenver this value is returned
+            # HostId is what it's called whenever this value is returned
             # in an XML response body, so to be consistent, we'll always
             # call is HostId.
             metadata['HostId'] = headers.get('x-amz-id-2', '')
@@ -753,9 +846,7 @@ class BaseRestParser(ResponseParser):
             payload_member_name = shape.serialization['payload']
             body_shape = member_shapes[payload_member_name]
             if body_shape.serialization.get('eventstream'):
-                parser = self._event_stream_parser
-                name = response['context'].get('operation_name')
-                body = EventStream(response['body'], body_shape, parser, name)
+                body = self._create_event_stream(response, body_shape)
                 final_parsed[payload_member_name] = body
             elif body_shape.type_name in ['string', 'blob']:
                 # This is a stream
@@ -853,7 +944,7 @@ class RestXMLParser(BaseRestParser, BaseXMLResponseParser):
 
     def _initial_body_parse(self, xml_string):
         if not xml_string:
-            return xml.etree.cElementTree.Element('')
+            return ETree.Element('')
         return self._parse_xml_string_to_dom(xml_string)
 
     def _do_error_parse(self, response, shape):
